#include "main.h"
#include "mqtt.h"
#include "timespec.h"
#include "frozen/frozen.h"

static void channel_timespec_cb(void *arg);
static void channel_timespec(int idx);

static struct channel_t s_channels[CHANNEL_MAX];
static int s_num_channels = 0;

static bool valid_gpio(const int gpio) {
  if (gpio == -1) {
    return true;
  }
  if (gpio < GPIO_MIN || gpio > GPIO_MAX) {
    return false;
  }
  return true;
}

int channel_get_total() {
  return s_num_channels;
}

struct mgos_timespec *channel_get_timespec(int idx) {
  if (idx < 0 || idx >= channel_get_total()) {
    return false;
  }

  return s_channels[idx].timespec;
}

bool channel_init(const char *fn) {
  char *            json;
  void *            h = NULL;
  struct json_token val;
  bool ret = false;

  char *name             = NULL;
  int   statusled        = -1;
  bool  statusled_invert = false;

  int idx;

  memset(s_channels, -1, sizeof(s_channels));
  s_num_channels = 0;

  json = json_fread(fn);
  if (!json) {
    LOG(LL_ERROR, ("%s: Could not json_fread()", fn));
    return false;
  }

  if (json_scanf(json, strlen(json), "{name:%Q, statusled:%d}", &name, &statusled) != 2) {
    LOG(LL_ERROR, ("Incomplete JSON: require 'name' and 'statusled' fields"));
    goto exit;
  }
  LOG(LL_INFO, ("Configuration: name='%s', statusled=%d", name, statusled));
  if (!valid_gpio(statusled)) {
    LOG(LL_ERROR, ("Status LED GPIO (%d) out of bounds [%d,%d]", statusled, GPIO_MIN, GPIO_MAX));
    goto exit;
  }
  json_scanf(json, strlen(json), "{statusled_invert:%B}", &statusled_invert);
  if (statusled_invert) {
    LOG(LL_INFO, ("Inverting statusled GPIO (%d)", statusled));
  }

  // Traverse Array
  while ((h = json_next_elem(json, strlen(json), h, ".channels", &idx, &val)) != NULL) {
    int  led = -1, relay = -1, button = -1;
    bool led_invert   = false;
    bool relay_invert = false;
    char fn[100];

    LOG(LL_DEBUG, ("[%d]: [%.*s]", idx, val.len, val.ptr));
    if (val.len == 0 || !val.ptr) {
      continue;
    }
    if (json_scanf(val.ptr, val.len, "{led:%d, relay:%d, button:%d}", &led, &relay, &button) != 3) {
      LOG(LL_ERROR, ("Incomplete Channel JSON: require 'led' and 'relay' and 'button' fields"));
      goto exit;
    }
    json_scanf(val.ptr, val.len, "{led:%d, led_invert:%B, relay:%d, relay_invert:%B, button:%d}", &led, &led_invert, &relay, &relay_invert, &button);
    if (!valid_gpio(led)) {
      LOG(LL_ERROR, ("LED GPIO (%d) out of bounds [%d,%d]", led, GPIO_MIN, GPIO_MAX));
      goto exit;
    }
    if (!valid_gpio(relay)) {
      LOG(LL_ERROR, ("Relay GPIO (%d) out of bounds [%d,%d]", relay, GPIO_MIN, GPIO_MAX));
      goto exit;
    }
    if (!valid_gpio(button)) {
      LOG(LL_ERROR, ("Button GPIO (%d) out of bounds [%d,%d]", button, GPIO_MIN, GPIO_MAX));
      goto exit;
    }
    if (idx == CHANNEL_MAX) {
      LOG(LL_ERROR, ("Too many channels (max is %d)", CHANNEL_MAX));
      goto exit;
    }
    LOG(LL_INFO, ("Channel[%d]: led=%d%s relay=%d%s button=%d ", idx, led, led_invert ? "(inverted)" : "", relay, relay_invert ? "(inverted)" : "", button));
    s_channels[idx].button_gpio      = button;
    s_channels[idx].relay_gpio       = relay;
    s_channels[idx].relay_invert     = relay_invert;
    s_channels[idx].led_gpio         = led;
    s_channels[idx].led_invert       = led_invert;
    s_channels[idx].channel_override = false;
    s_channels[idx].timespec         = timespec_create();

    snprintf(fn, sizeof(fn) - 1, "timespec.chan%d", idx);
    if (timespec_read_file(s_channels[idx].timespec, fn)) {
      LOG(LL_INFO, ("Initialized timespec from file %s", fn));
    }
  }

  ret = true;
exit:
  if (ret) {
    if (statusled != -1) {
      statusled_init(statusled, statusled_invert);
    }

    s_num_channels = idx + 1;
    LOG(LL_INFO, ("Configuring %d channels", s_num_channels));

    for (; idx >= 0; idx--) {
      s_channels[idx].relay_state = 0;

      if (s_channels[idx].relay_gpio != GPIO_INVALID) {
        mgos_gpio_set_mode(s_channels[idx].relay_gpio, MGOS_GPIO_MODE_OUTPUT);
        mgos_gpio_write(s_channels[idx].relay_gpio, (s_channels[idx].relay_state + s_channels[idx].relay_invert) % 2);
      }

      if (s_channels[idx].led_gpio != GPIO_INVALID) {
        mgos_gpio_set_mode(s_channels[idx].led_gpio, MGOS_GPIO_MODE_OUTPUT);
        mgos_gpio_write(s_channels[idx].led_gpio, (s_channels[idx].relay_state + s_channels[idx].led_invert) % 2);
      }

      if (s_channels[idx].button_gpio != GPIO_INVALID) {
        mgos_gpio_set_mode(s_channels[idx].button_gpio, MGOS_GPIO_MODE_INPUT);
        mgos_gpio_set_button_handler(s_channels[idx].button_gpio, MGOS_GPIO_PULL_UP, MGOS_GPIO_INT_EDGE_NEG, 20, channel_handler, (void *)1);
      }
    }
  }
  if (name) {
    free(name);
  }
  if (json) {
    free(json);
  }

  // Install timespec timer.
  mgos_set_timer(1000, true, channel_timespec_cb, NULL);
  return ret;
}

uint8_t channel_gpio_by_idx(int idx) {
  if (idx < 0 || idx >= s_num_channels) {
    return GPIO_INVALID;
  }
  return s_channels[idx].button_gpio;
}

uint8_t channel_idx_by_gpio(int gpio) {
  uint8_t i;

  for (i = 0; i < channel_get_total(); i++) {
    if (gpio == s_channels[i].button_gpio) {
      return i;
    }
  }
  return GPIO_INVALID;
}

struct channel_set_cb_args {
  uint8_t idx;
  bool    value;
};

static void channel_set_cb(void *arg) {
  struct channel_set_cb_args *cb_args = (struct channel_set_cb_args *)arg;

  if (!arg) {
    LOG(LL_ERROR, ("Callback arg is NULL"));
    return;
  }
  LOG(LL_INFO, ("Channel callback for channel %d value %d", cb_args->idx, cb_args->value));
  channel_set(cb_args->idx, cb_args->value);
  channel_override_set(cb_args->idx);
  free(arg);
}

void channel_set_duration(int idx, bool state, uint16_t seconds) {
  struct channel_set_cb_args *cb_args = calloc(1, sizeof(struct channel_set_cb_args));

  if (!cb_args) {
    LOG(LL_ERROR, ("Could not calloc() the callback args struct, aborting"));
    return;
  }

  channel_set(idx, state);
  channel_override_set(idx);

  // Set a timer to call back with the new intended state on the channel.
  cb_args->value = !state;
  cb_args->idx   = idx;
  LOG(LL_INFO, ("Setting timer for %d seconds on channel %d with future value %d", seconds, cb_args->idx, cb_args->value));
  mgos_set_timer(1000 * seconds, false /* oneshot */, channel_set_cb, cb_args);
}

void channel_set(int idx, bool state) {
  double now = mg_time();

  if (idx < 0 || idx >= channel_get_total()) {
    return;
  }

  s_channels[idx].button_last_change = now;
  s_channels[idx].relay_state        = state;
  if (s_channels[idx].relay_gpio != GPIO_INVALID) {
    mgos_gpio_write(s_channels[idx].relay_gpio, (s_channels[idx].relay_state + s_channels[idx].relay_invert) % 2);
  }
  if (s_channels[idx].led_gpio != GPIO_INVALID) {
    mgos_gpio_write(s_channels[idx].led_gpio, (s_channels[idx].relay_state + s_channels[idx].led_invert) % 2);
  }

  mqtt_publish_stat("channel", "{idx: %d, relay_state: %d}", idx, channel_get(idx));

  if (state) {
    // Sleep 250ms after toggling a relay to avoid cheaper boards from
    // suffering a brownout and rebooting the MCU in case several channels
    // are switched in short succession.
    mgos_msleep(250);
  }
}

bool channel_get(int idx) {
  if (idx < 0 || idx >= channel_get_total()) {
    return false;
  }

  return s_channels[idx].relay_state == 1;
}

// This method is called every second. It looks at the timespec for the channel
// and compares it with the channel_override flag, to see if the channel should
// be switched. Logic:
// - if timespec is missing or empty, return.
// - if channel_override is false: follow timespec and set channel accordingly.
// - if channel_override ie true:
//   - if timespec wants the channel set differently, leave it as-is
//   - if timespec agrees with the current setting, clear channel_override
// This behavior lets users set the channel (RPC or Button), and timespec will
// pick it up and govern the channel after the current (RPC/Button) state
// converges with the timespec.
static void channel_timespec(int idx) {
  bool match;

  if (idx < 0 || idx >= channel_get_total()) {
    return;
  }

  if (timespec_empty(s_channels[idx].timespec)) {
    return;
  }

  match = timespec_match_now(s_channels[idx].timespec);
  if (!s_channels[idx].channel_override) {
    // Follow timespec
    if (match == s_channels[idx].relay_state) {
      return;
    }
    LOG(LL_INFO, ("Timespec drives channel %d to relay_state %d", idx, match));
    mqtt_publish_stat("timespec", "{idx: %d, relay_state: %d, override: %B}", idx, match, s_channels[idx].channel_override);
    channel_set(idx, match);
  } else {
    // Follow override until it matches timespec
    if (match == s_channels[idx].relay_state) {
      LOG(LL_INFO, ("User state agrees with timespec, clearing override"));
      s_channels[idx].channel_override = false;
      mqtt_publish_stat("override", "{idx: %d, relay_state: %d, override: %B}", idx, match, s_channels[idx].channel_override);
      return;
    }
  }

  return;
}

static void channel_timespec_cb(void *arg) {
  uint8_t    i;
  time_t     now;
  struct tm *tm;

  now = time(NULL);
  if (now < 1541340332) {
    LOG(LL_WARN, ("NTP not yet synced, skipping."));
    return;
  }

  tm = localtime(&now);
  if (!tm) {
    LOG(LL_ERROR, ("Could not convert time (%u) into localtime", (uint32_t)now));
    return;
  }

  // LOG(LL_INFO, ("Time %u is %02d:%02d:%02d", (uint32_t) now, tm->tm_hour, tm->tm_min, tm->tm_sec));
  for (i = 0; i < channel_get_total(); i++) {
    channel_timespec(i);
  }
  (void)arg;
  return;
}

// Set an override flag if the user intended for a different state than the
// timespec. Do nothing otherwise.
void channel_override_set(int idx) {
  bool match;

  if (idx < 0 || idx >= channel_get_total()) {
    return;
  }

  if (timespec_empty(s_channels[idx].timespec)) {
    return;
  }

  match = timespec_match_now(s_channels[idx].timespec);
  if (match != s_channels[idx].relay_state) {
    LOG(LL_INFO, ("User state disagrees with timespec, setting override"));
    if (!s_channels[idx].channel_override) {
      s_channels[idx].channel_override = true;
      mqtt_publish_stat("override", "{idx: %d, relay_state: %d, override: %B}", idx, s_channels[idx].relay_state, s_channels[idx].channel_override);
    }
    return;
  }
  if (s_channels[idx].channel_override && (match == s_channels[idx].relay_state)) {
    LOG(LL_INFO, ("User state agrees with timespec, clearing override"));
    s_channels[idx].channel_override = false;
    mqtt_publish_stat("override", "{idx: %d, relay_state: %d, override: %B}", idx, s_channels[idx].relay_state, s_channels[idx].channel_override);
    return;
  }
}

// Clear the override flag on a channel.
void channel_override_clear(int idx) {
  if (idx < 0 || idx >= channel_get_total()) {
    return;
  }

  if (s_channels[idx].channel_override) {
    s_channels[idx].channel_override = false;
    mqtt_publish_stat("override", "{idx: %d, relay_state: %d, override: %B}", idx, s_channels[idx].relay_state, s_channels[idx].channel_override);
  }
  return;
}

// Channel handler arg is non-NULL if it was a pushbutton, and NULL if it was an RPC.
void channel_handler(int gpio, void *arg) {
  uint8_t idx;
  bool    state;

  idx = channel_idx_by_gpio(gpio);
  if (idx == GPIO_INVALID) {
    LOG(LL_ERROR, ("GPIO %d is not a valid button", gpio));
    return;
  }

  if (!arg) { // pushbutton
    LOG(LL_INFO, ("RPC triggered channel %d", idx));
  } else {
    LOG(LL_INFO, ("Button on GPIO %d triggered channel %d", gpio, idx));
    // Scan the button for 100ms to ensure it is still pressed. This is to avoid
    // RFI noise from triggering GPIO pins.
    for (int i = 0; i < 10; i++) {
      bool value;

      mgos_msleep(10);
      value = mgos_gpio_read(gpio);
      if (value) {
        LOG(LL_WARN, ("Button on GPIO %d no longer pressed after %d ms, skipping", gpio, 10 * i));
        return;
      }
    }
  }

  state = channel_get(idx);
  channel_set(idx, !state);
  channel_override_set(idx);
  (void)arg;
}
